#ifndef __At_Spatial_H__
#define __At_Spatial_H__

#include "../../../common/functions/functions.h"
#include "../../../common/filter/filter.h"

namespace Filtering { namespace AnisoTools { namespace Filters { namespace Spatial {

#define PIXTYPE Short
#define PRECISION 7
#define SPRECISION 15
#define STPRECISION 8

class Filter : public ChildFilter {

   PIXTYPE *pBuffer;
   Frame<PIXTYPE> frames[2];
   Frame<PIXTYPE> gradient;

   int *pEdges[3];
   int pSteps[3];
   int nIterations;
   int nCenter;

public:
   Filter(const Parameters &parameters) : ChildFilter(parameters), nIterations(parameters["iterations"].toInt()), nCenter(parameters["center"].toInt())
   {
      PIXTYPE *p = pBuffer = new PIXTYPE[3 * (nWidth + 2) * (nHeight + 2) * 3];
      frames[0] = Frame<PIXTYPE>(&p, nWidth, nHeight, C, 1, 1);
      frames[1] = Frame<PIXTYPE>(&p, nWidth, nHeight, C, 1, 1);
      gradient = Frame<PIXTYPE>(&p, nWidth, nHeight, C, 1, 1);

      for ( int i = 0; i < 3; i++ )
      {
         double edge = parameters["edge"].toFloat() * (1 << PRECISION);
         pEdges[i] = (new int[512 << PRECISION]) + (256 << PRECISION);
         for ( int j = -(256 << PRECISION); j < (256 << PRECISION); j++ )
            pEdges[i][j] = convert<int, Double>((1 << SPRECISION) * exp( -abs<Double>( j / edge ) ));
      }
      pSteps[0] = convert<int, Double>((1 << STPRECISION) * parameters["step"].toFloat());
      pSteps[1] = convert<int, Double>((1 << STPRECISION) * parameters["step"].toFloat());
      pSteps[2] = convert<int, Double>((1 << STPRECISION) * parameters["step"].toFloat());
   }
   ~Filter()
   {
      delete[] pBuffer;
      for ( int i = 0; i < 3; i++ )
         delete[] (pEdges[i] - (256 << PRECISION));
   }

   virtual Frame<Byte> get_frame(int n, const Frame<Byte> &output_frame);
   static Signature filter_signature()
   {
      Signature signature = "at_spatial";

      signature.add( Parameter( TYPE_CLIP, "" ) );
      signature.add( Parameter( 0.8f, "step" ) );
      signature.add( Parameter( 0.01f, "edge" ) );
      signature.add( Parameter( 5, "iterations" ) );
      signature.add( Parameter( 4, "center" ) );

      return signature;
   }

};

} } } } // namespaces Spatial, Filters, AnisoTools, Filtering

#endif // __At_Spatial_H__