#ifndef __At_Spatial_Functions_H__
#define __At_Spatial_Functions_H__

#include "../../common/frame/frame.h"

namespace Filtering { namespace AnisoTools { namespace Functions {

template<typename To, typename From> void ConvertPlane(const Plane<To> &to, const Plane<const From> &from, int shift)
{
   if ( shift > 0 )
      for ( int y = 0; y < from.height(); y++ )
         for ( int x = 0; x < from.width(); x++ )
            to[x + y * to.pitch()] = from[x + y * from.pitch() ] << shift;
   else if ( shift < 0 )
      for ( int y = 0; y < from.height(); y++ )
         for ( int x = 0; x < from.width(); x++ )
            to[x + y * to.pitch()] = convert<To, From>( (from[x + y * from.pitch() ] + (1 << (-shift - 1))) >> (-shift) );
   else
      for ( int y = 0; y < from.height(); y++ )
         for ( int x = 0; x < from.width(); x++ )
            to[x + y * to.pitch()] = convert<To, From>( from[x + y * from.pitch()] );
}

template<typename To, typename From> void ConvertFrame(const Frame<To> &to, const Frame<const From> &from, int shift)
{
   for ( int i = 0; i < plane_counts[to.colorspace()]; i++ )
      ConvertPlane<To, From>(to.plane(i), from.plane(i), shift);
}

template<typename Type> void PaddPlane(const Plane<Type> &plane, int nPaddingWidth, int nPaddingHeight)
{
   for ( int y = 0; y < nPaddingHeight; y++ )
   {
      for ( int x = 0; x < nPaddingWidth; x++ )
         plane[x + y * plane.pitch()] = plane[nPaddingWidth + nPaddingHeight * plane.pitch()];
      for ( int x = nPaddingWidth; x < plane.width() - nPaddingWidth; x++ )
         plane[x + y * plane.pitch()] = plane[x + nPaddingHeight * plane.pitch()];
      for ( int x = plane.width() - nPaddingWidth; x < plane.width(); x++ )
         plane[x + y * plane.pitch()] = plane[plane.width() - nPaddingWidth - 1 + nPaddingHeight * plane.pitch()];
   }
   for ( int y = nPaddingHeight; y < plane.height() - nPaddingHeight; y++ )
   {
      for ( int x = 0; x < nPaddingWidth; x++ )
         plane[x + y * plane.pitch()] = plane[nPaddingWidth + y * plane.pitch()];
      for ( int x = plane.width() - nPaddingWidth; x < plane.width(); x++ )
         plane[x + y * plane.pitch()] = plane[plane.width() - nPaddingWidth - 1 + y * plane.pitch()];
   }
   for ( int y = plane.height() - nPaddingHeight; y < plane.height(); y++ )
   {
      for ( int x = 0; x < nPaddingWidth; x++ )
         plane[x + y * plane.pitch()] = plane[nPaddingWidth + (plane.height() - nPaddingHeight - 1) * plane.pitch()];
      for ( int x = nPaddingWidth; x < plane.width() - nPaddingWidth; x++ )
         plane[x + y * plane.pitch()] = plane[x + (plane.height() - nPaddingHeight - 1) * plane.pitch()];
      for ( int x = plane.width() - nPaddingWidth; x < plane.width(); x++ )
         plane[x + y * plane.pitch()] = plane[plane.width() - nPaddingWidth - 1 + (plane.height() - nPaddingHeight - 1) * plane.pitch()];
   }
}

template<typename Type> void PaddFrame(const Frame<Type> &frame, int nPaddingWidth, int nPaddingHeight)
{
   for ( int i = 0; i < plane_counts[ frame.colorspace() ]; i++ )
      PaddPlane<Type>( frame.plane(i), nPaddingWidth, nPaddingHeight );
}

} } } // namespaces Functions, AnisoTools, Filtering

#endif // __At_Spatial_Functions_H__